//The VM emulation core, including a small disassembler / state dumper

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include "leds.h"
#include "main.h"
#include "buttons.h"
#include "motor.h"
#include "audio.h"
#include "update-mach.h"
#include "machine.h"

//All the storage of the VM
unsigned char mach_progmem[64*1024];
unsigned char mach_ram[256];
unsigned char mach_sources[64];
unsigned char mach_regs[16];
unsigned char mach_cmp1;
unsigned char mach_cmp2;
unsigned char mach_dc;
unsigned int mach_pc;
unsigned long mach_w;
int mach_input[16];

//The instruction type, as defined in the Violet vm asm manual, paragraph 4.8
//0=(inherent), 1=I, 2=R, 3=R,I, etc
//instype[first_byte_of_opcode] gives the type of that opcode.

int instype[]={
//  0 1 2 3 4 5 6 7 8 9 A B C D E F
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //LD
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //ADD
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //SUB
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //AND
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //OR
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //LDR
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3, //STR
    0,0,0,0,0,4,4,4,4,4,4,2,2,2,1,2, //7
    2,2,2,2,2,2,4,4,4,4,4,4,4,4,4,4, //8
    4,4,4,4,4,6,4,6,9,8,8,7,7,7,7,7, //9
    7,7,7,7,7,7,7,4,4,1,5,5,7,0,4,2, //A
    0,2,0,4,4,2,2,2,2,4,8,8,4,4,4,0, //B
    4,2,2,0,4,2,4,0,0,0,0,0,0,0,0,0, //C
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, //D
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, //E
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0  //F
	};

//The byte lengths of the various types of instructions
int inslen[]={1,2,2,2,2,3,3,3,4,4};

//mnemonics for instrs 0-0x60
char mnem_h[][4]={"LD","ADD","SUB","AND","OR","LDR","STR"};

//mnemonics for 60-end
char mnem_l[][12]={
    "NOP","na","RTI","CLRCC","SETCC","ADDCC","SUBCC","INCW",
    "DECW","MULW","INPUTRST","INT","na","WAIT","WAIT","RND",
    "DEC","INC","CLR","NEG","NOT","TST","LD","ADD",
    "SUB","MUL","AND","OR","EOR","LSL","LSR","ASR",
    "ROL","ROR","CMP","BIT","LDR","LDR","STR","STR",
    "LDT","LDTW","INPUT","RTIJ","BRA","BEQ","BNE","BGT",
    "BGE","BLT","BLE","BHI","BHS","BLO","BLS","LED",
    "PALETTE","na","PUSH","PULL","BSR","RTS","MOTOR","MIDIPLAY",
    "MIDISTOP","WAVPLAY","WAVSTOP","MSEC","SEC","BUT3","VOL","MVOL",
    "PUSHBUTTON","SRC","BRAT","BSRT","OSC","INV","DIV","HSV",
    "MOTORGET","MUSIC","DOWNLOAD","na","SEND","SENDREADY","LASTPING"};

//some handy redefinitions, to save some typing later :)
#define REG1 mach_regs[r1]
#define REG2 mach_regs[r2]
#define SP mach_regs[15]
#define cmp1 mach_cmp1
#define cmp2 mach_cmp2

//This routine dumps some info about the VMs current state
void dumpme() {
    int i;
    struct timeval tv;
    
    printf("Registers: ");
    for (i=0; i<16; i++) printf("R%i=%02X ",i,mach_regs[i]);
    gettimeofday(&tv,NULL);
    i=tv.tv_sec*1000+(tv.tv_usec/1000);
    printf("Sec: %04X Msec: %04X:",(int)(time(NULL)&0xffff),(int)((i)&0xffff));
    printf("w=%li\nInterrupts enabled:",mach_w);
    for (i=0; i<16; i++) {
	if (mach_input[i]!=-1) {
	    printf(" %i",i);
	    if (i==0) printf("-buttpress");
	    if (i==1) printf("-buttrelease");
	    if (i==2) printf("-vol");
	    if (i==14) printf("-invalid");
	    if (i==15) printf("-timer");
	}
    }	
    printf("\nMotor 1: pos=%i dir=%i   ",motorstate[0],motordir[0]);
    printf("Motor 2: pos=%i dir=%i\n",motorstate[1],motordir[1]);
}

//This routine simulates an hardware interrupt in the VM. i=interruptno.
void doint(int i) {
    if (mach_input[i]!=-1) {
	mach_ram[--SP]=mach_cmp2;
	mach_ram[--SP]=mach_cmp1;
	mach_ram[--SP]=mach_dc;
	mach_ram[--SP]=(mach_pc>>8);
	mach_ram[--SP]=mach_pc&0xff;
	mach_pc=mach_input[i];
    }
}

//This routine checks the hardware (buttons, volume, ...) for interrupt
//causes, and runs doint() if neccesary.
void check_interrupts() {
    static int oldmainbutton=0, oldvolume=0;
    static time_t oldtime=0;
    if (oldtime+5<time(NULL)) {
	doint(15);
	oldtime=time(NULL);
    }
    if (but_mainbutton!=oldmainbutton) {
	if (but_mainbutton==1) doint(0); else doint(1);
	oldmainbutton=but_mainbutton;
    }
    if (but_volume!=oldvolume) {
	doint(2);
	oldvolume=but_volume;
    }
}

//This ecodes the instruction at [addr] in program memory
int mach_vm_decode(long addr) {
    int ins,type;
    ins=mach_progmem[addr];

    if (ins<0x70) {
	printf(mnem_h[ins>>4]);
    } else {
	if (ins-0x70<88) {
	    printf(mnem_l[ins-0x70]);
	} else {
	    printf("na");
	}
    }
    type=instype[ins];
    if (type==0) {
	;
    } else if (type==1) {
	printf(" 0x%X",mach_progmem[addr+1]);
    } else if (type==2) {
	printf(" R%i",mach_progmem[addr+1]&0xf);
    } else if (type==3) {
	printf(" R%i,0x%X",mach_progmem[addr]&0xf,mach_progmem[addr+1]);
    } else if (type==4) {
	printf(" R%i,R%i",mach_progmem[addr+1]>>4,mach_progmem[addr+1]&0xf);
    } else if (type==5) {
	printf(" 0x%X,0x%X",mach_progmem[addr+1],mach_progmem[addr+2]);
    } else if (type==6) {
	printf(" R%i,0x%X,R%i",mach_progmem[addr+1]>>4,mach_progmem[addr+2],mach_progmem[addr+1]&0xf);
    } else if (type==7) {
	printf(" 0x%X",(mach_progmem[addr+1]<<8)+(mach_progmem[addr+2]));
    } else if (type==8) {
	printf(" R%i,0x%X",mach_progmem[addr+1],(mach_progmem[addr+2]<<8)+(mach_progmem[addr+3]));
    } else if (type==9) {
	printf(" R%i,R%i,0x%X",(mach_progmem[addr+1]>>4),mach_progmem[addr+1]&0xf,(mach_progmem[addr+2]<<8)+(mach_progmem[addr+3]));
    }
    return inslen[type];
}


//This resets the virtual machine
void mach_vm_reset() {
    int i;
    for (i=0; i<16; i++) mach_input[i]=-1;
    mach_pc=17;
    mach_w=0;
    mach_dc=0;
    mach_cmp1=0;
    mach_cmp2=0;
}


int mach_vm_run_till_wait() {
    long i=1000000;
    int t;
    t=mach_vm_step();
    while (mach_w==0 && i!=0 && t==0) {
	t=mach_vm_step();
	i--;
    }
    return t;
}

//This routine executes one instruction in the VM.
//If it return value !=0, something weird has happened (invalid instruction,
//...) and the main routine stops execution of the program.
int mach_vm_step() {
    int type;
    int r1=0,r2=0,i1=0,i2=0,ins;
    unsigned int i,j;
    float f, p, q, t;
    int h,s,v;
    long w=0;
    struct timeval tv; //for gettimeofday()

    check_interrupts();

    if (mach_w>0) {
	mach_w--;
	return 0;	
    }
    
    //Check instruction type and pre-fill some variables
    //with the operands so they're easily accessible later
    type=instype[mach_progmem[mach_pc]];
    if (type==1) {
	i1=mach_progmem[mach_pc+1];
    } else if (type==2) {
	r1=mach_progmem[mach_pc+1]&0xf;
    } else if (type==3) {
	r1=mach_progmem[mach_pc]&0xf;
	i1=mach_progmem[mach_pc+1];
    } else if (type==4) {
	r1=mach_progmem[mach_pc+1]>>4;
	r2=mach_progmem[mach_pc+1]&0xf;
    } else if (type==5) {
	i1=mach_progmem[mach_pc+1];
	i2=mach_progmem[mach_pc+2];
    } else if (type==6) {
	r1=mach_progmem[mach_pc+1]>>4;
	i1=mach_progmem[mach_pc+2];
	r2=mach_progmem[mach_pc+1]&0xf;
    } else if (type==7) {
	w=(mach_progmem[mach_pc+1]<<8)+(mach_progmem[mach_pc+2]);
    } else if (type==8) {
	r1=mach_progmem[mach_pc+1];
	w=(mach_progmem[mach_pc+2]<<8)+(mach_progmem[mach_pc+3]);
    } else if (type==9) {
	r1=(mach_progmem[mach_pc+1]>>4);
	r2=mach_progmem[mach_pc+1]&0xf;
	w=(mach_progmem[mach_pc+2]<<8)+(mach_progmem[mach_pc+3]);
    }
    ins=mach_progmem[mach_pc];
    //Increase pc to the start of the next instruction
    mach_pc+=inslen[instype[mach_progmem[mach_pc]]];
    
    //Handle all the one-byte R-instructions
    switch(ins&0xf0) {
    case 0:
	REG1=i1;
	cmp1=i1; cmp2=0;
	break;
    case 0x10:
	REG1+=i1;
	cmp1=REG1; cmp2=0;
	break;
    case 0x20:
	REG1-=i1;
	cmp1=REG1; cmp2=0;
	break;
    case 0x30:
	REG1&=i1;
	cmp1=REG1; cmp2=0;
	break;
    case 0x40:
	REG1|=i1;
	cmp1=REG1; cmp2=0;
	break;
    case 0x50:
	REG1=mach_ram[i1];
	cmp1=REG1; cmp2=0;
	break;
    case 0x60:
	mach_ram[i1]=REG1;
	cmp1=REG1; cmp2=0;
	break;
    default:
    //Handle the rest of the instructions
    //Watch out, while the indentation suggests this and the former switch-statement
    //are separate, they actually are nested.
    switch(ins) {
    case 0x70: //NOP
	//Well, let's do nothing :P
	;
        break;
    case 0x72: //RTI
	mach_pc=mach_ram[SP++];
	mach_pc+=(mach_ram[SP++]<<8);
	mach_dc=mach_ram[SP++];
	mach_cmp1=mach_ram[SP++];
	mach_cmp2=mach_ram[SP++];
        break;
    case 0x73: //CLRCC
	mach_dc=0;
        break;
    case 0x74: //SETCC
	mach_dc=1;
        break;
    case 0x75: //ADDCC
	REG2=REG2+REG1+mach_dc;
	cmp1=REG2; cmp2=0;
        break;
    case 0x76: //SUBCC
	REG2=REG2-REG1-mach_dc;
	cmp1=REG2; cmp2=0;
        break;
    case 0x77: //INCW
	i=(REG1<<8)+REG2;
	i++;
	REG1=(i>>8);
	REG2=(i&0xFF);
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0x78: //DECW
	i=(REG1<<8)+REG2;
	i++;
	REG1=(i>>8);
	REG2=(i&0xFF);
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0x79: //MULW
	i=REG1*REG2;
	REG1=i>>8;
	REG2=i&0xff;
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0x7A: //INPUTRST
	printf("TODO: inputrst\n");
//	return 1;
        break;
    case 0x7B: //INT
	mach_ram[--SP]=mach_cmp2;
	mach_ram[--SP]=mach_cmp1;
	mach_ram[--SP]=mach_dc;
	mach_ram[--SP]=(mach_pc>>8);
	mach_ram[--SP]=mach_pc&255;
	mach_pc=mach_input[REG1];
        break;
    case 0x7d: //WAIT
	mach_w=REG1;
        break;
    case 0x7e: //WAIT
	mach_w=i1;
        break;
    case 0x7f: //RND
	REG1=rand()&0xff;
        break;
    case 0x80: //DEC
	REG1--;
	cmp1=REG1; cmp2=0;
        break;
    case 0x81: //INC
	REG1++;
	cmp1=REG1; cmp2=0;
        break;
    case 0x82: //CLR
	REG1=0;
	cmp1=REG1; cmp2=0;
        break;
    case 0x83: //NEG
	REG1=-REG1;
	cmp1=REG1; cmp2=0;
        break;
    case 0x84: //NOT
	REG1=~REG1;
	cmp1=REG1; cmp2=0;
        break;
    case 0x85: //TST
	cmp1=REG1; cmp2=0;
        break;
    case 0x86: //LD
	REG1=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x87: //ADD
	REG1+=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x88: //SUB
	REG1-=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x89: //MUL
	i=REG1*REG2;
	REG1=(i&0xff);
	cmp1=REG1; cmp2=0;
        break;
    case 0x8A: //AND
	REG1&=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x8B: //OR
	REG1|=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x8C: //EOR
	REG1^=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x8D: //LSL
	REG1<<=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x8E: //LSR
	REG1>>=REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x8F: //ASR
	REG1=(char)REG1>>REG2;
	cmp1=REG1; cmp2=0;
        break;
    case 0x90: //ROL
	i=REG1<<REG2;
	REG1=(i&0xff)+(i>>8);
	cmp1=REG1; cmp2=0;
        break;
    case 0x91: //ROR
	i=(REG1<<8)>>REG2;
	REG1=(i&0xff)+(i>>8);
	cmp1=REG1; cmp2=0;
        break;
    case 0x92: //CMP
	cmp1=REG1; cmp2=REG2;
        break;
    case 0x93: //BIT
	cmp1=REG1&REG2; cmp2=0;
        break;
    case 0x94: //LDR
	REG1=mach_ram[REG2];
	cmp1=REG1; cmp2=0;
        break;
    case 0x95: //LDR
	REG1=mach_ram[REG2+i1];
	cmp1=REG1; cmp2=0;
        break;
    case 0x96: //STR
	mach_ram[REG2]=REG1;
	cmp1=REG1; cmp2=0;
        break;
    case 0x97: //STR
	mach_ram[REG2+i1]=REG1;
	cmp1=REG1; cmp2=0;
        break;
    case 0x98: //LDT
	REG1=mach_progmem[REG2+w];
	cmp1=REG1; cmp2=0;
        break;
    case 0x99: //LDTW
	REG1=mach_progmem[w+(mach_regs[3]<<8)+mach_regs[4]];
	cmp1=REG1; cmp2=0;
        break;
    case 0x9A: //INPUT
	mach_input[REG1]=w;
        break;
    case 0x9B: //RTIJ
	mach_pc=mach_ram[SP++];
	mach_pc|=mach_ram[SP++]<<8;
	mach_dc=mach_ram[SP++];
	mach_cmp1=mach_ram[SP++];
	mach_cmp2=mach_ram[SP++];
	mach_pc=w;
        break;
    case 0x9C: //BRA
	mach_pc=w;
        break;
    case 0x9D: //BEQ
	if (cmp1==cmp2) mach_pc=w;
        break;
    case 0x9E: //BNE
	if (cmp1!=cmp2) mach_pc=w;
        break;
    case 0x9F: //BGT
	if ((char)cmp1>(char)cmp2) mach_pc=w;
        break;
    case 0xA0: //BGE
	if ((char)cmp1>=(char)cmp2) mach_pc=w;
        break;
    case 0xA1: //BLT
	if ((char)cmp1<(char)cmp2) mach_pc=w;
        break;
    case 0xA2: //BLE
	if ((char)cmp1<=(char)cmp2) mach_pc=w;
        break;
    case 0xA3: //BHI
	if (cmp1>cmp2) mach_pc=w;
        break;
    case 0xA4: //BHS
	if (cmp1>=cmp2) mach_pc=w;
        break;
    case 0xA5: //BLO
	if (cmp1<cmp2) mach_pc=w;
        break;
    case 0xA6: //BLS
	if (cmp1<=cmp2) mach_pc=w;
        break;
    case 0xA7: //LED
	led_fade_to(REG1, mach_regs[0], mach_regs[1], mach_regs[2], REG2);
        break;
    case 0xA8: //PALETTE
	i=REG1;
	if (i&1) mach_regs[0]=REG2; else mach_regs[0]=0;
	if (i&2) mach_regs[1]=REG2; else mach_regs[1]=0;
	if (i&4) mach_regs[2]=REG2; else mach_regs[2]=0;
        break;
    case 0xA9: //Mysterious undocumented instruction. No idea what this does,
		//we'll have to figure it out on the hardware itself.
	printf("Mystery-instruction encountered.\n");
//	return 1;
	break;
    case 0xAA: //PUSH
	i=(i1<<8)+i2;
	if (i&0x8000) mach_ram[--SP]=mach_dc;
	i<<=1;
	for (j=0; j<15; j++) {
	    if (i&0x8000) mach_ram[--SP]=mach_regs[j];
	    i<<=1;
	}
        break;
    case 0xAB: //PULL
	i=(i1<<8)+i2;
	for (j=15; j>0; j--) {
	    if (i&1) mach_regs[j-1]=mach_ram[SP++];
	    i>>=1;
	}
	if (i&1) mach_dc=mach_ram[SP++];
        break;
    case 0xAC: //BSR
	mach_ram[--SP]=mach_pc>>8;
	mach_ram[--SP]=mach_pc&0xff;
	mach_pc=w;
        break;
    case 0xAD: //RTS
	mach_pc=mach_ram[SP++];
	mach_pc|=mach_ram[SP++]<<8;
        break;
    case 0xAE: //MOTOR
	if (REG1<2) {
		motordir[REG1]=REG2;
	}
        break;
    case 0xAF: //MIDIPLAY
	playmidi(REG1);
        break;
    case 0xB0: //MIDISTOP
	stopmidi();
        break;
    case 0xB1: //WAVPLAY
	playpcm(REG1);
        break;
    case 0xB2: //WAVSTOP
	stoppcm();
        break;
    case 0xB3: //MSEC
	gettimeofday(&tv,NULL);
	i=tv.tv_sec*1000+(tv.tv_usec/1000);
	REG1=(i>>8)&0xff;
	REG2=i&0xff;
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0xB4: //SEC
	i=time(NULL);
	REG1=(i>>8)&0xff;
	REG2=i&0xff;
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0xB5: //BUT3
	REG1=but_volume;
	cmp1=REG1; cmp2=0;
        break;
    case 0xB6: //VOL
	printf("ToDo: vol\n");
//	return 1;
        break;
    case 0xB7: //MVOL
	printf("ToDo: mvol\n");
//	return 1;
        break;
    case 0xB8: //PUSHBUTTON
	REG1=but_mainbutton;
	cmp1=REG1; cmp2=0;
        break;
    case 0xB9: //SRC
	REG1=mach_sources[REG2];
	cmp1=REG1; cmp2=0;
        break;
    case 0xBA: //BRAT
	i=w+(REG1*2);
	mach_pc=(mach_progmem[i]<<8)+mach_progmem[i+1];
        break;
    case 0xBB: //BSRT
	mach_ram[--SP]=mach_pc>>8;
	mach_ram[--SP]=mach_pc&0xff;
	i=w+(REG1*2);
	mach_pc=(mach_progmem[i]<<8)+mach_progmem[i+1];
        break;
    case 0xBC: //OSC
	REG1=(unsigned char)(128*((float)1-cos((float)REG2*3.141592/128)));
	cmp1=REG1; cmp2=0;
        break;
    case 0xBD: //INV
	i=65536/REG2;
	REG1=(i>>8);
	REG2=(i&0xFF);
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0xBE: //DIV
	if (REG2==0) {
	    i=65535;
	} else {
	    i=(256*REG1)/REG2;
	}
	REG1=(i>>8);
	REG2=(i&0xFF);
	cmp1=REG1|REG2; cmp2=0;
        break;
    case 0xBF: //HSV2RGB
	//Joinked from http://www.cs.rit.edu/~ncs/color/t_convert.html
	//Hope it still works :)
	h=mach_regs[0]; s=mach_regs[1]; v=mach_regs[2];
	if( s == 0 ) {
	    // achromatic (grey)
	    mach_regs[0] = mach_regs[1] = mach_regs[2] = v;
	} else {
	    h /= 60;			// sector 0 to 5
	    i = floor( h );
	    f = h - i;			// factorial part of h
	    p = v * ( 1 - s );
	    q = v * ( 1 - s * f );
	    t = v * ( 1 - s * ( 1 - f ) );

	    if (i==0) mach_regs[0]=v; mach_regs[1]=t; mach_regs[2]=p;
	    if (i==1) mach_regs[0]=q; mach_regs[1]=v; mach_regs[2]=p;
	    if (i==2) mach_regs[0]=p; mach_regs[1]=v; mach_regs[2]=t;
	    if (i==3) mach_regs[0]=p; mach_regs[1]=q; mach_regs[2]=v;
	    if (i==4) mach_regs[0]=t; mach_regs[1]=p; mach_regs[2]=v;
	    if (i==5) mach_regs[0]=v; mach_regs[1]=p; mach_regs[2]=q;
	}
        break;
    case 0xC0: //MOTORGET
	if (REG2<2) REG1=motorcounter[REG2];
        break;
    case 0xC1: //MUSIC
	printf("ToDo: music\n");
	REG1=0;
//	return 1;
        break;
    case 0xC2: //DOWNLOAD
	printf("ToDo: download\n");
	REG1=0;
//	return 1;
        break;
    case 0xC4: //SEND
	send_do((REG1<<8)+REG2);
	return 1;
        break;
    case 0xC5: //SENDREADY
	printf("ToDo: sendready\n");
	REG1=0;
//	return 1;
        break;
    case 0xC6: //LASTPING
	i=time(NULL)-lastupdated;
	REG1=(i>>8);
	REG2=(i&255);
	cmp1=REG1; cmp2=REG2;
	break;
    default:
	printf("INVALID OPCODE 0x%X!\n",ins);
	printf("PC=%04x ",mach_pc);
	return 2;
	break;
    }
    }
    return 0;
}


